# Imports do próprio módulo

from inewave.config import (
    MAX_ANOS_ESTUDO,
    MAX_ANOS_HISTORICO,
    MAX_CONFIGURACOES,
    MESES_ABREV,
    MAX_REES,
    MESES_DF,
)

# Imports de módulos externos
from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from datetime import date
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from typing import IO, List
from inewave._utils.formatacao import (
    prepara_vetor_anos_tabela,
    repete_vetor,
)


class BlocoSerieEnergiaREE(Block):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às séries de energia por REE.
    """

    __slots__ = [
        "__campo_ree",
        "__campo_cfg",
        "__linha",
    ]

    BEGIN_PATTERN = "SERIE  DE ENERGIAS DO REE"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_ree = LiteralField(16, 55)
        self.__campo_cfg = IntegerField(5, 88)
        campo_ano: List[Field] = [IntegerField(4, 0)]
        campo_energia: List[Field] = [
            FloatField(10, 4 + 11 * i, 2) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campo_energia)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoSerieEnergiaREE):
            return False
        bloco: BlocoSerieEnergiaREE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "data": prepara_vetor_anos_tabela(anos),
                    "valor": tabela.flatten(),
                }
            )
            df["ree"] = ree
            df["configuracao"] = cfg
            return df[["ree", "configuracao", "data", "valor"]]

        # Identifica o REE e a configuração
        linha = file.readline()
        ree = self.__campo_ree.read(linha)
        cfg = self.__campo_cfg.read(linha)

        # Salta as linhas adicionais
        for _ in range(4):
            file.readline()

        # Variáveis auxiliares
        anos: List[int] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO, len(MESES_DF)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            anos.append(dados[0])
            tabela[i, :] = dados[1:]
            i += 1


class BlocoCorrelEnergiasREE(Block):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às séries de autocorrelações
    das energias por REE.
    """

    __slots__ = [
        "__linha",
    ]

    BEGIN_PATTERN = "CORRELOGRAMO DA SERIE DE ENERGIAS"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_mes: List[Field] = [LiteralField(3, 1)]
        campo_ano: List[Field] = [LiteralField(4, 5)]
        correl: List[Field] = [
            FloatField(10, 9 + 10 * i, 5) for i in range(11)
        ]
        self.__linha = Line(campo_mes + campo_ano + correl)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelEnergiasREE):
            return False
        bloco: BlocoCorrelEnergiasREE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // 12
            numero_anos_pos = len([p for p in anos if p == "POS"]) // 12
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = sorted(anos_estudo)[0]
            # Descobre o último ano de estudo
            ultimo_ano_estudo = sorted(anos_estudo)[-1]
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx_i = 12 * a
                idx_f = idx_i + 12
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * 12
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + 12 * a
                idx_f = idx_i + 12
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * 12
            return [int(a) for a in anos]

        def converte_vetor_meses(meses: List[str]) -> List[int]:
            return [MESES_ABREV.index(m) + 1 for m in meses]

        def converte_tabela_em_df():
            anos_conv = converte_vetor_anos(anos)
            meses_conv = converte_vetor_meses(meses)
            datas = [
                date(year=a, month=m, day=1)
                for a, m in zip(anos_conv, meses_conv)
            ]
            df = pd.DataFrame(
                data={
                    "data": repete_vetor(datas, len(MESES_DF) - 1),
                    "lag": np.tile(np.arange(1, len(MESES_DF)), len(datas)),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(5):
            file.readline()

        # Variáveis auxiliares
        meses: List[str] = []
        anos: List[str] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO * len(MESES_DF), 11))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            meses.append(dados[0])
            anos.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1


class BlocoCorrelParcialEnergiasREE(Block):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às séries de autocorrelações parciais
    das energias por REE.
    """

    __slots__ = [
        "__linha",
    ]

    BEGIN_PATTERN = "CORRELOGRAMO PARCIAL DA SERIE DE ENERGIAS"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_mes: List[Field] = [LiteralField(3, 1)]
        campo_ano: List[Field] = [LiteralField(4, 5)]
        correl: List[Field] = [
            FloatField(10, 9 + 10 * i, 5) for i in range(11)
        ]
        self.__linha = Line(campo_mes + campo_ano + correl)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelParcialEnergiasREE):
            return False
        bloco: BlocoCorrelParcialEnergiasREE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // 12
            numero_anos_pos = len([p for p in anos if p == "POS"]) // 12
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = sorted(anos_estudo)[0]
            # Descobre o último ano de estudo
            ultimo_ano_estudo = sorted(anos_estudo)[-1]
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx_i = 12 * a
                idx_f = idx_i + 12
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * 12
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + 12 * a
                idx_f = idx_i + 12
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * 12
            return [int(a) for a in anos]

        def converte_vetor_meses(meses: List[str]) -> List[int]:
            return [MESES_ABREV.index(m) + 1 for m in meses]

        def converte_tabela_em_df():
            anos_conv = converte_vetor_anos(anos)
            meses_conv = converte_vetor_meses(meses)
            datas = [
                date(year=a, month=m, day=1)
                for a, m in zip(anos_conv, meses_conv)
            ]
            df = pd.DataFrame(
                data={
                    "data": repete_vetor(datas, len(MESES_DF) - 1),
                    "lag": np.tile(np.arange(1, len(MESES_DF)), len(datas)),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            file.readline()

        # Variáveis auxiliares
        meses: List[str] = []
        anos: List[str] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO * len(MESES_DF), 11))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            meses.append(dados[0])
            anos.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1


class BlocoOrdemModeloREE(Block):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às ordens finais ajustadas do modelo e aos
    seus coeficientes por REE.
    """

    __slots__ = [
        "__linha",
        "__tipo",
    ]

    BEGIN_PATTERN = "DO MODELO AUTORREGRESSIVO PARA CADA PERIODO"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_ano: List[Field] = [LiteralField(4, 32)]
        orders: List[Field] = [
            IntegerField(5, 36 + 5 * i) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + orders)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoOrdemModeloREE):
            return False
        bloco: BlocoOrdemModeloREE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"])
            numero_anos_pos = len([p for p in anos if p == "POS"])
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = sorted(anos_estudo)[0]
            # Descobre o último ano de estudo
            ultimo_ano_estudo = sorted(anos_estudo)[-1]
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx = a
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx] = str(ano)
            for a in range(numero_anos_pos):
                idx = indice_inicio_pos + a
                ano = ultimo_ano_estudo + a + 1
                anos[idx] = str(ano)
            return [int(a) for a in anos]

        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "data": prepara_vetor_anos_tabela(
                        converte_vetor_anos(anos)
                    ),
                    "valor": tabela.flatten(),
                }
            )
            df["tipo"] = self.__tipo
            return df[["tipo", "data", "valor"]]

        # Salta as linhas adicionais
        linha = file.readline()
        self.__tipo = linha.split("ORDEM")[1].split("DO MODELO")[0].strip()
        for _ in range(3):
            file.readline()

        # Variáveis auxiliares
        anos: List[str] = []
        tabela = np.zeros((3 * MAX_ANOS_ESTUDO, len(MESES_DF)), dtype=np.int64)
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            anos.append(dados[0])
            tabela[i, :] = dados[1:]
            i += 1


class BlocoCoeficientesModeloREE(Block):
    """
    Lista de coeficientes dos modelos PAR ou PAR-A por REE.
    """

    __slots__ = [
        "__linha",
    ]

    BEGIN_PATTERN = " COEFICIENTES DA EQUACAO DE REGRESSAO DE UM PROCESSO"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        coefs: List[Field] = [
            FloatField(9, 11 * i, 3, format="E") for i in range(11)
        ]
        self.__linha = Line(coefs)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCoeficientesModeloREE):
            return False
        bloco: BlocoCoeficientesModeloREE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "tipo": repete_vetor(["psi", "psi_norm"], len(MESES_DF)),
                    "ordem": np.tile(list(range(1, len(MESES_DF))) + ["A"], 2),
                    "valor": tabela.flatten(),
                }
            )
            return df.dropna()

        linha = file.readline()

        tabela = np.zeros((1, 2 * len(MESES_DF)))
        # Processa os dados
        dados_psi = self.__linha.read(file.readline())
        dados_psi_norm = self.__linha.read(file.readline())
        linha = file.readline()
        if len(linha) < 3:
            dados_psi_a = [0.0, 0.0]
        else:
            dados_psi_a = [
                self.__linha.read(linha)[0],
                self.__linha.read(file.readline())[0],
            ]
        tabela[0, :] = (
            dados_psi + [dados_psi_a[0]] + dados_psi_norm + [dados_psi_a[1]]
        )
        self.data = converte_tabela_em_df()


class BlocoSerieRuidosREE(Block):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às séries de ruídos por REE.
    """

    __slots__ = [
        "__campo_ano",
        "__linha",
    ]

    BEGIN_PATTERN = "SERIE DE RUIDOS  - ANO:"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_ano = LiteralField(5, 81)
        campos: List[Field] = [
            FloatField(11, 3 + 11 * i, 3, format="E")
            for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campos)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoSerieRuidosREE):
            return False
        bloco: BlocoSerieRuidosREE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "valor": tabela.flatten(),
                }
            )
            df["ano"] = ano
            df["mes"] = np.tile(
                np.arange(1, len(MESES_DF) + 1), tabela.shape[0]
            )
            df["serie"] = repete_vetor(list(range(1, i + 1)))
            return df[["ano", "serie", "mes", "valor"]]

        # Identifica o ano em questão
        linha = file.readline()
        ano = self.__campo_ano.read(linha)

        # Salta as linhas adicionais
        for _ in range(5):
            file.readline()

        # Variáveis auxiliares
        tabela = np.zeros((MAX_ANOS_HISTORICO, len(MESES_DF)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            tabela[i, :] = self.__linha.read(linha)
            i += 1


class BlocoCorrelRuidosREE(Block):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às séries de autocorrelações
    dos ruídos por REE.
    """

    __slots__ = [
        "__linha",
    ]

    BEGIN_PATTERN = "CORRELOGRAMO DA SERIE DE RUIDOS"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_mes: List[Field] = [LiteralField(3, 1)]
        campo_ano: List[Field] = [LiteralField(4, 5)]
        correl: List[Field] = [
            FloatField(10, 9 + 10 * i, 5) for i in range(11)
        ]
        self.__linha = Line(campo_mes + campo_ano + correl)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelRuidosREE):
            return False
        bloco: BlocoCorrelRuidosREE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // 12
            numero_anos_pos = len([p for p in anos if p == "POS"]) // 12
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = sorted(anos_estudo)[0]
            # Descobre o último ano de estudo
            ultimo_ano_estudo = sorted(anos_estudo)[-1]
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx_i = 12 * a
                idx_f = idx_i + 12
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * 12
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + 12 * a
                idx_f = idx_i + 12
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * 12
            return [int(a) for a in anos]

        def converte_vetor_meses(meses: List[str]) -> List[int]:
            return [MESES_ABREV.index(m) + 1 for m in meses]

        def converte_tabela_em_df():
            anos_conv = converte_vetor_anos(anos)
            meses_conv = converte_vetor_meses(meses)
            datas = [
                date(year=a, month=m, day=1)
                for a, m in zip(anos_conv, meses_conv)
            ]
            df = pd.DataFrame(
                data={
                    "data": repete_vetor(datas, len(MESES_DF) - 1),
                    "lag": np.tile(np.arange(1, len(MESES_DF)), len(datas)),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(4):
            file.readline()

        # Variáveis auxiliares
        meses: List[str] = []
        anos: List[str] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO * len(MESES_DF), 11))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            meses.append(dados[0])
            anos.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1


class BlocoSerieMediasREE(Block):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às séries de médias de 12 meses por REE.
    """

    __slots__ = [
        "__campo_ano",
        "__linha",
    ]

    BEGIN_PATTERN = "SERIE MEDIA 12 MESES - ANO:"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_ano = LiteralField(5, 77)
        campos: List[Field] = [
            FloatField(11, 3 + 11 * i, 2, format="F")
            for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campos)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoSerieMediasREE):
            return False
        bloco: BlocoSerieMediasREE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "valor": tabela.flatten(),
                }
            )
            df["ano"] = ano
            df["mes"] = np.tile(
                np.arange(1, len(MESES_DF) + 1), tabela.shape[0]
            )
            df["serie"] = repete_vetor(list(range(1, i + 1)))
            return df[["ano", "serie", "mes", "valor"]]

        # Identifica o ano em questão
        linha = file.readline()
        ano = self.__campo_ano.read(linha)

        # Salta as linhas adicionais
        for _ in range(5):
            file.readline()

        # Variáveis auxiliares
        tabela = np.zeros((MAX_ANOS_HISTORICO, len(MESES_DF)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            tabela[i, :] = self.__linha.read(linha)
            i += 1


class BlocoCorrelCruzadaMediaREE(Block):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às séries de correlação cruzada da variável
    anual com as energias por REE.
    """

    __slots__ = [
        "__linha",
    ]

    BEGIN_PATTERN = "CORRELACAO CRUZADA VARIAVEL ANUAL  ENERGIAS"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_mes: List[Field] = [LiteralField(3, 1)]
        campo_ano: List[Field] = [LiteralField(4, 5)]
        correl: List[Field] = [
            FloatField(10, 9 + 10 * i, 5) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_mes + campo_ano + correl)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelCruzadaMediaREE):
            return False
        bloco: BlocoCorrelCruzadaMediaREE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // 12
            numero_anos_pos = len([p for p in anos if p == "POS"]) // 12
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = sorted(anos_estudo)[0]
            # Descobre o último ano de estudo
            ultimo_ano_estudo = sorted(anos_estudo)[-1]
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx_i = 12 * a
                idx_f = idx_i + 12
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * 12
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + 12 * a
                idx_f = idx_i + 12
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * 12
            return [int(a) for a in anos]

        def converte_vetor_meses(meses: List[str]) -> List[int]:
            return [MESES_ABREV.index(m) + 1 for m in meses]

        def converte_tabela_em_df():
            anos_conv = converte_vetor_anos(anos)
            meses_conv = converte_vetor_meses(meses)
            datas = [
                date(year=a, month=m, day=1)
                for a, m in zip(anos_conv, meses_conv)
            ]
            df = pd.DataFrame(
                data={
                    "data": repete_vetor(datas, len(MESES_DF)),
                    "lag": np.tile(
                        np.arange(1, len(MESES_DF) + 1), len(datas)
                    ),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(5):
            file.readline()

        # Variáveis auxiliares
        meses: List[str] = []
        anos: List[str] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO * len(MESES_DF), len(MESES_DF)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            meses.append(dados[0])
            anos.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1


class BlocoCorrelEspacialAnualConfig(Block):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo à correlação espacial histórica espacial
    anual por configuração.
    """

    __slots__ = ["__campo_cfg", "__linha"]

    BEGIN_PATTERN = "CORRELACAO ESPACIAL HISTORICA ANUAL"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_cfg = IntegerField(4, 68)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelEspacialAnualConfig):
            return False
        bloco: BlocoCorrelEspacialAnualConfig = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=rees)
            df["ree"] = rees
            df["configuracao"] = cfg
            df = df[["configuracao", "ree"] + rees]
            return df

        # Identifica a configuração
        linha = file.readline()
        cfg = self.__campo_cfg.read(linha)

        file.readline()
        linha = file.readline()
        rees = [r.strip() for r in linha.split(" ") if len(r) > 2]
        campo_ree: List[Field] = [LiteralField(12, 0)]
        campo_correl: List[Field] = [
            FloatField(13, 12 + 13 * i, 4) for i in range(len(rees))
        ]
        self.__linha = Line(campo_ree + campo_correl)

        # Variáveis auxiliares
        tabela = np.zeros((len(rees), len(rees)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            tabela[i, :] = self.__linha.read(linha)[1:]
            i += 1


class BlocoCorrelEspacialMensalConfig(Block):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo à correlação espacial histórica espacial
    mensal por configuração.
    """

    __slots__ = ["__campo_cfg", "__linha"]

    BEGIN_PATTERN = "CORRELACAO ESPACIAL HISTORICA MENSAL"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_cfg = IntegerField(4, 69)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelEspacialMensalConfig):
            return False
        bloco: BlocoCorrelEspacialMensalConfig = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=rees)
            df["ree"] = valores_rees
            df["configuracao"] = cfg
            df = df[["configuracao", "ree"] + rees]
            df = df.rename(columns={"MES": "mes"})
            df = df.astype({"mes": "int64"})
            return df

        # Identifica a configuração
        linha = file.readline()
        cfg = self.__campo_cfg.read(linha)

        file.readline()
        linha = file.readline()
        rees = [r.strip() for r in linha.split(" ") if len(r) > 2]
        campo_ree: List[Field] = [LiteralField(12, 0)]
        campo_mes: List[Field] = [IntegerField(2, 15)]
        campo_correl: List[Field] = [
            FloatField(8, 17 + 13 * i, 4) for i in range(len(rees) - 1)
        ]
        self.__linha = Line(campo_ree + campo_mes + campo_correl)

        # Variáveis auxiliares
        tabela = np.zeros(
            (MAX_CONFIGURACOES * MAX_REES * len(MESES_DF), len(rees))
        )
        valores_rees: List[str] = []
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                linha = file.readline()
                if len(linha) < 4:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                    break
                else:
                    continue
            # Senão, processa os dados
            dados_linha = self.__linha.read(linha)
            valores_rees.append(dados_linha[0])
            tabela[i, :] = dados_linha[1:]
            i += 1



from inewave.newave.modelos.conft import BlocoConfUTE

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Conft(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às configurações das
    usinas térmicas.
    """

    T = TypeVar("T")

    SECTIONS = [BlocoConfUTE]

    @property
    def usinas(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as usinas.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - submercado (`int`)
        - usina_existente (`str`)
        - classe (`int`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoConfUTE)
        if isinstance(b, BlocoConfUTE):
            return b.data
        return None

    @usinas.setter
    def usinas(self, valor: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoConfUTE)
        if isinstance(b, BlocoConfUTE):
            b.data = valor
        else:
            raise ValueError("Campo não lido")



# Inclui os membros

# Deprecated
from .avl_cortesfpha_nwv import AvlCortesFpha  # noqa
from .avl_desvfpha_s import AvlDesvFphaS  # noqa
from .avl_desvfpha_v_q import AvlDesvFphaVQ  # noqa
from .eco_fpha import EcoFpha  # noqa
from .nwv_avl_evap import NwvAvlEvap  # noqa
from .nwv_cortes_evap import NwvCortesEvap  # noqa
from .nwv_eco_evap import NwvEcoEvap  # noqa

from .abertura import Abertura  # noqa
from .adterm import Adterm  # noqa
from .agrint import Agrint  # noqa
from .arquivos import Arquivos  # noqa
from .bid import BID  # noqa
from .cadic import Cadic  # noqa
from .caso import Caso  # noqa
from .clasgas import Clasgas  # noqa
from .clast import Clast  # noqa
from .confhd import Confhd  # noqa
from .conft import Conft  # noqa
from .curva import Curva  # noqa
from .cvar import Cvar  # noqa
from .dger import Dger  # noqa
from .dsvagua import Dsvagua  # noqa
from .eafpast import Eafpast  # noqa
from .elnino import ElNino  # noqa
from .ensoaux import ENSOAux  # noqa
from .exph import Exph  # noqa
from .expt import Expt  # noqa
from .gee import GEE  # noqa
from .ghmin import Ghmin  # noqa
from .gtminpat import Gtminpat  # noqa
from .hidr import Hidr  # noqa
from .itaipu import Itaipu  # noqa
from .manutt import Manutt  # noqa
from .modif import Modif  # noqa
from .newavetim import Newavetim  # noqa
from .parp import Parp  # noqa
from .parpvaz import Parpvaz  # noqa
from .parpeol import Parpeol  # noqa
from .patamar import Patamar  # noqa
from .penalid import Penalid  # noqa
from .perda import Perda  # noqa
from .pmo import Pmo  # noqa
from .re import Re  # noqa
from .ree import Ree  # noqa
from .sar import Sar  # noqa
from .shist import Shist  # noqa
from .selcor import Selcor  # noqa
from .sistema import Sistema  # noqa
from .tecno import Tecno  # noqa
from .term import Term  # noqa
from .vazoes import Vazoes  # noqa
from .vazpast import Vazpast  # noqa
from .volref_saz import VolrefSaz  # noqa
from .fpha_cortes import FphaCortes  # noqa
from .fpha_avl_desv_s import FphaAvlDesvS  # noqa
from .fpha_avl_desv_v_q import FphaAvlDesvVQ  # noqa
from .fpha_eco import FphaEco  # noqa
from .evap_avl_desv import EvapAvlDesv  # noqa
from .evap_cortes import EvapCortes  # noqa
from .evap_eco import EvapEco  # noqa
from .energiaf import Energiaf  # noqa
from .energiab import Energiab  # noqa
from .energias import Energias  # noqa
from .enavazf import Enavazf  # noqa
from .enavazb import Enavazb  # noqa
from .engnat import Engnat  # noqa
from .vazaof import Vazaof  # noqa
from .vazaob import Vazaob  # noqa
from .vazaos import Vazaos  # noqa
from .vazinat import Vazinat  # noqa



from inewave.nwlistop.modelos.nwlistopdat import (
    BlocoDadosNwlistop,
)

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional, List


class Nwlistopdat(SectionFile):
    """
    Armazena os dados de entrada para a execução do programa auxiliar
    NWLISTOP, existentes no arquivo `nwlistop.dat`.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoDadosNwlistop]

    @property
    def opcao(self) -> Optional[int]:
        """
        A opção de execução do programa nwlistop.

        :return: O flag de opção
        :rtype: Optional[int] | None
        """
        b = self.data.get_sections_of_type(BlocoDadosNwlistop)
        if isinstance(b, BlocoDadosNwlistop):
            return b.data.get("opcao")
        return None

    @opcao.setter
    def opcao(self, v: int):
        b = self.data.get_sections_of_type(BlocoDadosNwlistop)
        if isinstance(b, BlocoDadosNwlistop):
            b.data["opcao"] = v

    @property
    def periodo_inicial_impressao(self) -> Optional[int]:
        """
        O período inicial para impressão dos dados de saída.

        :return: O índice do período
        :rtype: Optional[int] | None
        """
        if self.opcao in [1, 2, 4]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                return b.data["periodos"][0]
            return None
        else:
            raise ValueError("Períodos só são suportados nas opções [1, 2, 4]")

    @periodo_inicial_impressao.setter
    def periodo_inicial_impressao(self, v: int):
        if self.opcao in [1, 2, 4]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                b.data["periodos"][0] = v
        else:
            raise ValueError("Períodos só são suportados nas opções [1, 2, 4]")

    @property
    def periodo_final_impressao(self) -> Optional[int]:
        """
        O período final para impressão dos dados de saída.

        :return: O índice do período
        :rtype: Optional[int] | None
        """
        if self.opcao in [1, 2, 4]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                return b.data["periodos"][1]
            return None
        else:
            raise ValueError("Períodos só são suportados nas opções [1, 2, 4]")

    @periodo_final_impressao.setter
    def periodo_final_impressao(self, v: int):
        if self.opcao in [1, 2, 4]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                b.data["periodos"][1] = v
        else:
            raise ValueError("Períodos só são suportados nas opções [1, 2, 4]")

    @property
    def serie_inicial_impressao(self) -> Optional[int]:
        """
        A série inicial para impressão dos dados de saída.

        :return: O índice da série
        :rtype: Optional[int] | None
        """
        if self.opcao in [1]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                return b.data["series"][0]
            return None
        else:
            raise ValueError("Séries só são suportadas na opção [1]")

    @serie_inicial_impressao.setter
    def serie_inicial_impressao(self, v: int):
        if self.opcao in [1]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                b.data["series"][0] = v
        else:
            raise ValueError("Séries só são suportadas na opção [1]")

    @property
    def serie_final_impressao(self) -> Optional[int]:
        """
        A série final para impressão dos dados de saída.

        :return: O índice da série
        :rtype: Optional[int] | None
        """
        if self.opcao in [1]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                return b.data["series"][1]
            return None
        else:
            raise ValueError("Séries só são suportadas na opção [1]")

    @serie_final_impressao.setter
    def serie_final_impressao(self, v: int):
        if self.opcao in [1]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                b.data["series"][1] = v
        else:
            raise ValueError("Séries só são suportadas na opção [1]")

    @property
    def variaveis_impressao_estagios_agregados(self) -> Optional[List[int]]:
        """
        As variáveis dos períodos agregados a serem impressas na
        opção 2.

        :return: A lista dos códigos das variáveis.
        :rtype: Optional[List[int]] | None
        """
        if self.opcao in [2]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                return b.data["variaveis_ree"]
            return None
        else:
            raise ValueError("Variáveis só são suportadas na opção [2]")

    @variaveis_impressao_estagios_agregados.setter
    def variaveis_impressao_estagios_agregados(self, v: List[int]):
        if self.opcao in [2]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                b.data["variaveis_ree"] = v
        else:
            raise ValueError("Variáveis só são suportadas na opção [2]")

    @property
    def variaveis_impressao_estagios_individualizados(
        self,
    ) -> Optional[List[int]]:
        """
        As variáveis dos períodos individualizados a serem impressas na
        opção 2.

        :return: A lista dos códigos das variáveis.
        :rtype: Optional[List[int]] | None
        """
        if self.opcao in [2]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                return b.data["variaveis_uhe"]
            return None
        else:
            raise ValueError("Variáveis só são suportadas na opção [2]")

    @variaveis_impressao_estagios_individualizados.setter
    def variaveis_impressao_estagios_individualizados(self, v: List[int]):
        if self.opcao in [2]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                b.data["variaveis_uhe"] = v
        else:
            raise ValueError("Variáveis só são suportadas na opção [2]")

    @property
    def uhes_impressao_estagios_individualizados(
        self,
    ) -> Optional[List[int]]:
        """
        As UHEs que terão as variáveis impressas na opção 2.

        :return: A lista dos códigos das UHEs.
        :rtype: Optional[List[int]] | None
        """
        if self.opcao in [2]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                return b.data["uhes"]
            return None
        else:
            raise ValueError("UHEs só são suportadas na opção [2]")

    @uhes_impressao_estagios_individualizados.setter
    def uhes_impressao_estagios_individualizados(self, v: List[int]):
        if self.opcao in [2]:
            b = self.data.get_sections_of_type(BlocoDadosNwlistop)
            if isinstance(b, BlocoDadosNwlistop):
                b.data["uhes"] = v
        else:
            raise ValueError("UHEs só são suportadas na opção [2]")



# Rotinas de testes associadas ao arquivo MEDIAS-RHV.CSV do NWLISTOP
from inewave.nwlistop.mediasrhv import Mediasrhv

ARQ_TESTE = "tests/_arquivos/MEDIAS-RHV.CSV"


def test_eq_mediasrhv():
    leitor = Mediasrhv.read(ARQ_TESTE)
    leitor2 = Mediasrhv.read(ARQ_TESTE)
    assert leitor == leitor2
    assert leitor.valores.shape == (0, 2)



from inewave.newave.avl_desvfpha_v_q import AvlDesvFphaVQ
import pandas as pd  # type: ignore
from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.avl_desvfpha_v_q import MockAvlDesvFphaVQ

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_atributos_encontrados_avl_desvfpha_v_q():
    m: MagicMock = mock_open(read_data="".join(MockAvlDesvFphaVQ))
    with patch("builtins.open", m):
        rel = AvlDesvFphaVQ.read(ARQ_TESTE)
        assert rel.versao == "FPHA_NEWAVE"
        assert rel.tabela.at[0, "codigo_usina"] == 4
        assert rel.tabela.at[0, "nome_usina"] == "FUNIL-GRANDE"
        assert pd.isna(rel.tabela.at[0, "volume_armazenado_percentual"])
        assert rel.tabela.at[0, "vazao_turbinada_m3s"] == 0.0
        assert rel.tabela.at[0, "desvio_percentual"] == 0.0


def test_eq_avl_desvfpha_v_q():
    m: MagicMock = mock_open(read_data="".join(MockAvlDesvFphaVQ))
    with patch("builtins.open", m):
        rel1 = AvlDesvFphaVQ.read(ARQ_TESTE)
        rel2 = AvlDesvFphaVQ.read(ARQ_TESTE)
        assert rel1 == rel2


## Não precisa comparar isto, é somente leitura - passa por agregação
# def test_neq_avl_desvfpha_v_q():
#     m: MagicMock = mock_open(read_data="".join(MockAvlDesvFphaVQ))
#     with patch("builtins.open", m):
#         rel1 = AvlDesvFphaVQ.read(ARQ_TESTE)
#         rel2 = AvlDesvFphaVQ.read(ARQ_TESTE)
#         rel1.tabela.iloc[0, 0] = -1
#         assert rel1 != rel2



from inewave.nwlistop.modelos.blocos.restricao import Restricao
from inewave.nwlistop.modelos.arquivos.arquivorestricao import (
    ArquivoRestricao,
)
from inewave.nwlistop.modelos.form_rhv import FormRHVAnos


class FormRHV(ArquivoRestricao):
    """
    Armazena os dados das saídas referentes aos valores das restrições
    RHV por restrição.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `form_rhvXXX.out`.

    """

    BLOCKS = [
        Restricao,
        FormRHVAnos,
    ]



from inewave.config import MESES_DF

from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField

from inewave.nwlistop.modelos.blocos.valoresclassetermicaseriepatamar import (
    ValoresClasseTermicaSeriePatamar,  # type: ignore
)


class GTAnos(ValoresClasseTermicaSeriePatamar):
    """
    Bloco com as informações das tabelas de geração térmica por classe.
    """

    __slots__ = []

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [  # type: ignore
            IntegerField(3, 2),
            IntegerField(4, 7),
            LiteralField(1, 15),
        ]
        + [FloatField(9, 16 + 9 * i, 1) for i in range(len(MESES_DF))]  # type: ignore
    )



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)
from inewave.nwlistop.modelos.vevminsin import VevminAnos

from warnings import warn


class Vevminsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes às violações da meta
    de energia da vazão mínima para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `vevminsin.out`

    """

    BLOCKS = [
        VevminAnos,
    ]

    def __init__(self, data=...) -> None:
        warn(
            "Esta classe é relativa a um arquivo que não é mais suportado."
            + " Utilize a classe ViolEvminsin no lugar.",
            DeprecationWarning,
        )
        super().__init__(data)



from typing import TypeVar, List, Optional, Union, Type
from datetime import datetime
import pandas as pd  # type: ignore
from cfinterface.components.register import Register
from cfinterface.files.registerfile import RegisterFile
from inewave.libs.modelos.eolica import (
    RegistroPEECadastro,
    RegistroPEEPotenciaInstaladaPeriodo,
    RegistroPEEConfiguracaoPeriodo,
    RegistroPEEFTE,
    RegistroPEEGeracaoPatamar,
    RegistroHistoricoVentoHorizonte,
    RegistroHistoricoVento,
    RegistroPEEPostoVento,
    RegistroPostoVentoCadastro,
    RegistroPEESubmercado,
)


class Eolica(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao dados
    das usinas eólicas do sistema.
    """

    T = TypeVar("T", bound=Register)

    REGISTERS = [
        RegistroPEECadastro,
        RegistroPEEPotenciaInstaladaPeriodo,
        RegistroPEEConfiguracaoPeriodo,
        RegistroPEEFTE,
        RegistroPEEGeracaoPatamar,
        RegistroHistoricoVentoHorizonte,
        RegistroHistoricoVento,
        RegistroPEEPostoVento,
        RegistroPostoVentoCadastro,
        RegistroPEESubmercado,
    ]

    def __registros_ou_df(
        self, t: Type[T], **kwargs
    ) -> Optional[Union[T, List[T], pd.DataFrame]]:
        if kwargs.get("df"):
            return self._as_df(t)
        else:
            kwargs_sem_df = {k: v for k, v in kwargs.items() if k != "df"}
            return self.data.get_registers_of_type(t, **kwargs_sem_df)

    def pee_cad(
        self,
        codigo_pee: Optional[int] = None,
        nome_pee: Optional[str] = None,
        df: bool = False,
    ) -> Optional[
        Union[RegistroPEECadastro, List[RegistroPEECadastro], pd.DataFrame]
    ]:
        """
        Obtém um registro que cadastra um PEE.

        :param codigo_pee: código que especifica o PEE
        :type codigo_pee: int | None
        :param nome_pee: nome do PEE
        :type nome_pee: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEECadastro` |
            list[:class:`RegistroPEECadastro`] | None
        """
        return self.__registros_ou_df(
            RegistroPEECadastro,
            codigo_pee=codigo_pee,
            nome_pee=nome_pee,
            df=df,
        )

    def pee_pot_inst_per(
        self,
        codigo_pee: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        potencia_instalada: Optional[float] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            RegistroPEEPotenciaInstaladaPeriodo,
            List[RegistroPEEPotenciaInstaladaPeriodo],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém um registro que contém a potência instalada por período
        para um PEE.

        :param codigo_pee: código que especifica o PEE
        :type codigo_pee: int | None
        :param data_inicio: período de início de operação
        :type data_inicio: datetime | None
        :param data_fim: período de fim de operação
        :type data_fim: datetime | None
        :param potencia_instalada: potência efetiva do PEE
        :type potencia_instalada: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEEPotenciaInstaladaPeriodo` |
            list[:class:`RegistroPEEPotenciaInstaladaPeriodo`] | None
        """
        return self.__registros_ou_df(
            RegistroPEEPotenciaInstaladaPeriodo,
            codigo_pee=codigo_pee,
            data_inicio=data_inicio,
            data_fim=data_fim,
            potencia_instalada=potencia_instalada,
            df=df,
        )

    def pee_config_per(
        self,
        codigo_pee: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        estado_operacao: Optional[str] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            RegistroPEEConfiguracaoPeriodo,
            List[RegistroPEEConfiguracaoPeriodo],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém um registro que contém a configuração de operação para
        um PEE em um período.

        :param codigo_pee: código que especifica o PEE
        :type codigo_pee: int | None
        :param data_inicio: período de início da configuração
        :type data_inicio: datetime | None
        :param data_fim: período de fim da configuração
        :type data_fim: datetime | None
        :param estado_operacao: tipo de operação
        :type estado_operacao: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEEConfiguracaoPeriodo` |
            list[:class:`RegistroPEEConfiguracaoPeriodo`] | None
        """
        return self.__registros_ou_df(
            RegistroPEEConfiguracaoPeriodo,
            codigo_pee=codigo_pee,
            data_inicio=data_inicio,
            data_fim=data_fim,
            estado_operacao=estado_operacao,
            df=df,
        )

    def pee_fpvp_lin_pu_per(
        self,
        codigo_pee: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        coeficiente_linear: Optional[float] = None,
        coeficiente_angular: Optional[float] = None,
        df: bool = False,
    ) -> Optional[Union[RegistroPEEFTE, List[RegistroPEEFTE], pd.DataFrame]]:
        """
        Obtém um registro que contém a função de produção vento-geração
        para um período de tempo para um PEE, em p.u.

        :param codigo_pee: código que especifica o PEE
        :type codigo_pee: int | None
        :param data_inicio: a data inicial de validade para a função
        :type data_inicio: datetime | None
        :param data_fim: a data final de validade para a função
        :type data_fim: datetime | None
        :param coeficiente_linear: o coeficiente linear
        :type coeficiente_linear: float | None
        :param coeficiente_angular: o coeficiente angular
        :type coeficiente_angular: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEEFTE` |
            list[:class:`RegistroPEEFTE`] | None
        """
        return self.__registros_ou_df(
            RegistroPEEFTE,
            codigo_pee=codigo_pee,
            data_inicio=data_inicio,
            data_fim=data_fim,
            coeficiente_linear=coeficiente_linear,
            coeficiente_angular=coeficiente_angular,
            df=df,
        )

    def pee_ger_prof_per_pat(
        self,
        codigo_pee: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        indice_patamar: Optional[int] = None,
        profundidade: Optional[float] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            RegistroPEEGeracaoPatamar,
            List[RegistroPEEGeracaoPatamar],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém um registro que contém a profundidade de um patamar
        de geração para um período, para um PEE.

        :param codigo_pee: código do PEE
        :type codigo_pee: int | None
        :param data_inicio: data de início do registro histórico
        :type data_inicio: datetime | None
        :param data_fim: data de fim do registro histórico
        :type data_fim: datetime | None
        :param indice_patamar: patamar de geração eólica
        :type indice_patamar: int | None
        :param profundidade: profundidade do patamar de geração
        :type profundidade: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEEGeracaoPatamar` |
            list[:class:`RegistroPEEGeracaoPatamar`] | None
        """
        return self.__registros_ou_df(
            RegistroPEEGeracaoPatamar,
            codigo_pee=codigo_pee,
            data_inicio=data_inicio,
            data_fim=data_fim,
            indice_patamar=indice_patamar,
            profundidade=profundidade,
            df=df,
        )

    def vento_hist_horiz(
        self,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            RegistroHistoricoVentoHorizonte,
            List[RegistroHistoricoVentoHorizonte],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém um registro que contém o horizonte de um histórico.

        :param data_inicio: data de início do histórico
        :type data_inicio: datetime | None
        :param data_fim: data de fim do histórico
        :type data_fim: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroHistoricoVentoHorizonte` |
            list[:class:`RegistroHistoricoVentoHorizonte`] | None
        """
        return self.__registros_ou_df(
            RegistroHistoricoVentoHorizonte,
            data_inicio=data_inicio,
            data_fim=data_fim,
            df=df,
        )

    def vento_hist(
        self,
        codigo_eolica: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        velocidade: Optional[float] = None,
        direcao: Optional[float] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            RegistroHistoricoVento, List[RegistroHistoricoVento], pd.DataFrame
        ]
    ]:
        """
        Obtém um registro que contém a o valor de vento para um
        intervalo do histórico.

        :param codigo_eolica: código da usina eólica
        :type codigo_eolica: int | None
        :param data_inicio: data de início do registro histórico
        :type data_inicio: datetime | None
        :param data_fim: data de fim do registro histórico
        :type data_fim: datetime | None
        :param velocidade: velocidade do vento
        :type velocidade: float | None
        :param direcao: direção do vento
        :type direcao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroHistoricoVento` |
            list[:class:`RegistroHistoricoVento`] | None
        """
        return self.__registros_ou_df(
            RegistroHistoricoVento,
            codigo_eolica=codigo_eolica,
            data_inicio=data_inicio,
            data_fim=data_fim,
            velocidade=velocidade,
            direcao=direcao,
            df=df,
        )

    def posto_vento_cad(
        self,
        codigo_posto: Optional[int] = None,
        nome_posto: Optional[str] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            RegistroPostoVentoCadastro,
            List[RegistroPostoVentoCadastro],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém um registro que contém o cadastro de um posto
            de vento.

        :param codigo_posto: código que identifica o posto
        :type codigo_posto: int | None
        :param nome_posto: nome do posto de vento
        :type nome_posto: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPostoVentoCadastro` |
            list[:class:`RegistroPostoVentoCadastro`] | None
        """
        return self.__registros_ou_df(
            RegistroPostoVentoCadastro,
            codigo_posto=codigo_posto,
            nome_posto=nome_posto,
            df=df,
        )

    def pee_posto(
        self,
        codigo_pee: Optional[datetime] = None,
        codigo_posto: Optional[datetime] = None,
        df: bool = False,
    ) -> Optional[
        Union[RegistroPEEPostoVento, List[RegistroPEEPostoVento], pd.DataFrame]
    ]:
        """
        Obtém um registro que contém o mapeamento entre um posto
            e um PEE.

        :param codigo_pee: código que identifica o PEE
        :type codigo_pee: int | None
        :param codigo_posto: código que identifica o posto
        :type codigo_posto: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEEPostoVento` |
            list[:class:`RegistroPEEPostoVento`] | None
        """
        return self.__registros_ou_df(
            RegistroPEEPostoVento,
            codigo_pee=codigo_pee,
            codigo_posto=codigo_posto,
            df=df,
        )

    def pee_subm(
        self,
        codigo_pee: Optional[int] = None,
        codigo_submercado: Optional[int] = None,
        df: bool = False,
    ) -> Optional[
        Union[RegistroPEESubmercado, List[RegistroPEESubmercado], pd.DataFrame]
    ]:
        """
        Obtém um registro que contém o mapeamento PEE-submercado.

        :param codigo_pee: código que especifica o PEE
        :type codigo_pee: int | None
        :param codigo_submercado: código que especifica o submercado
        :type codigo_submercado: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEESubmercado` |
            list[:class:`RegistroPEESubmercado`] | None
        """
        return self.__registros_ou_df(
            RegistroPEESubmercado,
            codigo_pee=codigo_pee,
            codigo_submercado=codigo_submercado,
            df=df,
        )



# Rotinas de testes associadas ao arquivo volref_saz.dat do NEWAVE
from inewave.newave.modelos.volref_saz import BlocoVolrefSaz

from inewave.newave import VolrefSaz


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.volref_saz import MockVolrefSaz

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_bloco_desvios_volrefsaz():
    m: MagicMock = mock_open(read_data="".join(MockVolrefSaz))
    b = BlocoVolrefSaz()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 2040
    assert b.data.iloc[0, 0] == 4
    assert b.data.iloc[-1, -1] == 0.0


def test_atributos_encontrados_volrefsaz():
    m: MagicMock = mock_open(read_data="".join(MockVolrefSaz))
    with patch("builtins.open", m):
        ad = VolrefSaz.read(ARQ_TESTE)
        assert ad.volumes is not None


def test_atributos_nao_encontrados_volrefsaz():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = VolrefSaz.read(ARQ_TESTE)
        assert ad.volumes is None


def test_eq_volrefsaz():
    m: MagicMock = mock_open(read_data="".join(MockVolrefSaz))
    with patch("builtins.open", m):
        cf1 = VolrefSaz.read(ARQ_TESTE)
        cf2 = VolrefSaz.read(ARQ_TESTE)
        assert cf1 == cf2


def test_neq_volrefsaz():
    m: MagicMock = mock_open(read_data="".join(MockVolrefSaz))
    with patch("builtins.open", m):
        cf1 = VolrefSaz.read(ARQ_TESTE)
        cf2 = VolrefSaz.read(ARQ_TESTE)
        cf2.volumes.iloc[0, 0] = -1
        assert cf1 != cf2


def test_leitura_escrita_volrefsaz():
    m_leitura: MagicMock = mock_open(read_data="".join(MockVolrefSaz))
    with patch("builtins.open", m_leitura):
        cf1 = VolrefSaz.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = VolrefSaz.read(ARQ_TESTE)
        assert cf1 == cf2



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.floatfield import FloatField
from typing import IO
from datetime import datetime
from inewave.config import MAX_ANOS_HISTORICO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class SecaoDadosVazinat(Section):
    """
    Registro com os dados das séries históricas de vazão incremental por
    configuração existentes no arquivo vazinat.dat.
    """

    __slots__ = ["__linha"]

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, SecaoDadosVazinat):
            return False
        bloco: SecaoDadosVazinat = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    def read(
        self,
        file: IO,
        numero_uhes: int = 164,
        numero_configuracoes: int = 60,
        ano_inicio_historico: int = 1931,
        *args,
        **kwargs,
    ):
        numero_registros = (
            MAX_ANOS_HISTORICO * 12 * numero_uhes * numero_configuracoes
        )
        self.__linha = Line(
            [
                FloatField(size=8, starting_position=8 * i)
                for i in range(numero_registros)
            ],
            storage="BINARY",
        )
        dados = self.__linha.read(file.read(self.__linha.size))
        datas_df = pd.date_range(
            datetime(year=ano_inicio_historico, month=1, day=1),
            datetime(
                year=ano_inicio_historico + MAX_ANOS_HISTORICO - 1,
                month=12,
                day=1,
            ),
            freq="MS",
        ).to_numpy()
        datas_df = np.tile(datas_df, numero_uhes * numero_configuracoes)
        uhes_df = np.tile(
            np.repeat(np.arange(1, numero_uhes + 1), 12 * MAX_ANOS_HISTORICO),
            numero_configuracoes,
        )
        configuracoes_df = np.repeat(
            np.arange(1, numero_configuracoes + 1),
            12 * MAX_ANOS_HISTORICO * numero_uhes,
        )

        df = pd.DataFrame(
            data={
                "configuracao": configuracoes_df,
                "data": datas_df,
                "indice_usina": uhes_df,
                "valor": dados,
            }
        )
        self.data = df

    def write(self, file: IO, *args, **kwargs):
        dados = self.data["valor"].to_numpy()
        linha = Line(
            [
                FloatField(size=8, starting_position=8 * i)
                for i in range(len(dados))
            ],
            storage="BINARY",
        )
        file.write(linha.write(dados))



from inewave.nwlistcf.modelos.arquivos import BlocoNomesArquivos

from cfinterface.files.sectionfile import SectionFile

from typing import List, TypeVar, Optional
import pandas as pd  # type: ignore


class Arquivos(SectionFile):
    """
    Armazena os dados de entrada do NWLISTCF referentes ao arquivo
    `arquivos.dat`.

    Esta classe lida com informações de entrada do NWLISTCF e
    que deve se referir aos nomes dos demais arquivos de entrada
    utilizados para o caso em questão.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoNomesArquivos]

    def __le_nome_por_indice(self, indice: int) -> Optional[str]:
        b = self.data.get_sections_of_type(BlocoNomesArquivos)
        if isinstance(b, BlocoNomesArquivos):
            if indice in b.data.index:
                dado = b.data.iloc[indice, 1]
                if isinstance(dado, str):
                    return dado
        return None

    def __atualiza_nome_por_indice(self, indice: int, nome: str):
        b = self.data.get_sections_of_type(BlocoNomesArquivos)
        if isinstance(b, BlocoNomesArquivos):
            dif = indice - b.data.shape[0] + 1
            if dif > 0:
                col_vazia = [None] * dif
                b.data = pd.concat(
                    [
                        b.data,
                        pd.DataFrame(
                            data={
                                "Legenda": col_vazia,
                                "Nome": col_vazia,
                            }
                        ),
                    ],
                    ignore_index=True,
                )
            b.data.iloc[indice, 1] = nome

    @property
    def arquivos(self) -> List[str]:
        """
        Os nomes dos arquivos utilizados.

        :return: Os arquivos na mesma ordem em que são declarados
        :rtype: List[str]
        """
        b = self.data.get_sections_of_type(BlocoNomesArquivos)
        return (
            [] if not isinstance(b, BlocoNomesArquivos) else b.data.iloc[:, 1]
        )

    @property
    def nwlistcf(self) -> Optional[str]:
        """
        Nome do arquivo de dados gerais utilizado pelo NWLISTCF.
        """
        return self.__le_nome_por_indice(0)

    @nwlistcf.setter
    def nwlistcf(self, arq: str):
        self.__atualiza_nome_por_indice(0, arq)

    @property
    def cortes(self) -> Optional[str]:
        """
        Nome do arquivo de cortes gerado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(1)

    @cortes.setter
    def cortes(self, arq: str):
        self.__atualiza_nome_por_indice(1, arq)

    @property
    def cortesh(self) -> Optional[str]:
        """
        Nome do arquivo de cabeçalho de cortes produzido pelo NEWAVE.
        """
        return self.__le_nome_por_indice(2)

    @cortesh.setter
    def cortesh(self, arq: str):
        self.__atualiza_nome_por_indice(2, arq)

    @property
    def newdesp(self) -> Optional[str]:
        """
        Nome do arquivo para despacho hidrotérmico produzido pelo NEWAVE.
        """
        return self.__le_nome_por_indice(3)

    @newdesp.setter
    def newdesp(self, arq: str):
        self.__atualiza_nome_por_indice(3, arq)

    @property
    def cortese(self) -> Optional[str]:
        """
        Nome do arquivo dos estados dos cortes produzido pelo NEWAVE.
        """
        return self.__le_nome_por_indice(4)

    @cortese.setter
    def cortese(self, arq: str):
        self.__atualiza_nome_por_indice(4, arq)

    @property
    def energiaf(self) -> Optional[str]:
        """
        Nome do arquivo com as energias para simulação
        forward produzido pelo NEWAVE.
        """
        return self.__le_nome_por_indice(5)

    @energiaf.setter
    def energiaf(self, arq: str):
        self.__atualiza_nome_por_indice(5, arq)

    @property
    def rsar(self) -> Optional[str]:
        """
        Nome do arquivo com restrições SAR produzido pelo NEWAVE.
        """
        return self.__le_nome_por_indice(6)

    @rsar.setter
    def rsar(self, arq: str):
        self.__atualiza_nome_por_indice(6, arq)

    @property
    def rsarh(self) -> Optional[str]:
        """
        Nome do arquivo com cabeçalho da SAR produzido pelo NEWAVE.
        """
        return self.__le_nome_por_indice(7)

    @rsarh.setter
    def rsarh(self, arq: str):
        self.__atualiza_nome_por_indice(7, arq)

    @property
    def rsari(self) -> Optional[str]:
        """
        Nome do arquivo de índice da SAR produzido pelo NEWAVE.
        """
        return self.__le_nome_por_indice(8)

    @rsari.setter
    def rsari(self, arq: str):
        self.__atualiza_nome_por_indice(8, arq)

    @property
    def relatorio_cortes(self) -> Optional[str]:
        """
        Nome do arquivo com a saída do NWLISTCF contendo os
        cortes produzidos pelo NEWAVE.
        """
        return self.__le_nome_por_indice(9)

    @relatorio_cortes.setter
    def relatorio_cortes(self, arq: str):
        self.__atualiza_nome_por_indice(9, arq)

    @property
    def relatorio_estados(self) -> Optional[str]:
        """
        Nome do arquivo com a saída do NWLISTCF contendo os
        estados visitados pelo NEWAVE.
        """
        return self.__le_nome_por_indice(10)

    @relatorio_estados.setter
    def relatorio_estados(self, arq: str):
        self.__atualiza_nome_por_indice(10, arq)

    @property
    def relatorio_rsar(self) -> Optional[str]:
        """
        Nome do arquivo com a saída do NWLISTCF contendo os
        a SAR utilizada pelo NEWAVE.
        """
        return self.__le_nome_por_indice(11)

    @relatorio_rsar.setter
    def relatorio_rsar(self, arq: str):
        self.__atualiza_nome_por_indice(11, arq)

    @property
    def energiaxf(self) -> Optional[str]:
        """
        Nome do arquivo com as médias móveis das energias
        para simulação forward produzido pelo NEWAVE.
        """
        return self.__le_nome_por_indice(12)

    @energiaxf.setter
    def energiaxf(self, arq: str):
        self.__atualiza_nome_por_indice(12, arq)

    @property
    def vazaof(self) -> Optional[str]:
        """
        Nome do arquivo com as vazões para simulação
        forward produzido pelo NEWAVE.
        """
        return self.__le_nome_por_indice(13)

    @vazaof.setter
    def vazaof(self, arq: str):
        self.__atualiza_nome_por_indice(13, arq)

    @property
    def vazaoxf(self) -> Optional[str]:
        """
        Nome do arquivo com as médias móveis de vazões
        para simulação forward produzido pelo NEWAVE.
        """
        return self.__le_nome_por_indice(14)

    @vazaoxf.setter
    def vazaoxf(self, arq: str):
        self.__atualiza_nome_por_indice(14, arq)



from cfinterface.files.sectionfile import SectionFile
from inewave.newave.modelos.forwarh import SecaoDadosForwarh


from typing import TypeVar, Optional, List


class Forwarh(SectionFile):
    """
    Armazena os dados de saída do NEWAVE referentes ao
    cabeçalho dos dados das simulações forward.
    """

    T = TypeVar("T")

    SECTIONS = [SecaoDadosForwarh]
    STORAGE = "BINARY"

    def __bloco_dados(self) -> Optional[SecaoDadosForwarh]:
        dados = [r for r in self.data.of_type(SecaoDadosForwarh)]
        if len(dados) == 1:
            return dados[0]
        else:
            return None

    @property
    def nome_caso(self) -> Optional[str]:
        """
        O nome do caso, conforme cadastrado no `dger.dat`.

        :return: O nome do caso
        :rtype: str
        """
        dados = self.__bloco_dados()
        return dados.nome_caso if dados is not None else None

    @property
    def numero_rees(self) -> Optional[int]:
        """
        O número de REEs cadastrados no `ree.dat`.

        :return: O número de REEs
        :rtype: str
        """
        dados = self.__bloco_dados()
        return dados.numero_rees if dados is not None else None

    @property
    def numero_submercados(self) -> Optional[int]:
        """
        O número de submercados cadastrados no `sistema.dat`.

        :return: O número de submercados
        :rtype: str
        """
        dados = self.__bloco_dados()
        return dados.numero_submercados if dados is not None else None

    @property
    def numero_series_gravadas(self) -> Optional[int]:
        """
        O número de séries gravadas, configurado no `dger.dat`.

        :return: O número de séries gravadas
        :rtype: str
        """
        dados = self.__bloco_dados()
        return dados.numero_series_gravadas if dados is not None else None

    @property
    def numero_aberturas(self) -> Optional[int]:
        """
        O número de aberturas, configurado no `dger.dat`.

        :return: O número de aberturas
        :rtype: str
        """
        dados = self.__bloco_dados()
        return dados.numero_aberturas if dados is not None else None

    @property
    def numero_estagios_estudo(self) -> Optional[int]:
        """
        O número de estágios do período de estudo, configurado no `dger.dat`.

        :return: O número de estágios
        :rtype: str
        """
        dados = self.__bloco_dados()
        return dados.numero_estagios_estudo if dados is not None else None

    @property
    def intervalo_series_gravadas(self) -> Optional[int]:
        """
        O intervalo de espaçamento das séries gravadas, configurado no `dger.dat`.

        :return: O intervalo entre as séries
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.intervalo_series_gravadas if dados is not None else None

    @property
    def numero_classes_termicas_submercados(self) -> Optional[List[int]]:
        """
        O número de classes térmicas por submercado, configurado no `clast.dat`.

        :return: A lista com o número de classes térmicas
        :rtype: list[int] | None
        """
        dados = self.__bloco_dados()
        return (
            dados.numero_classes_termicas_submercados
            if dados is not None
            else None
        )

    @property
    def numero_patamares_deficit(self) -> Optional[int]:
        """
        O número de patamares de déficit, configurado no `sistema.dat`.

        :return: O número de patamares
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.numero_patamares_deficit if dados is not None else None

    @property
    def tamanho_registro_arquivo_forward(self) -> Optional[int]:
        """
        O tamanho de um registro do arquivo forward, em bytes.

        :return: O número de bytes de um registro
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return (
            dados.tamanho_registro_arquivo_forward
            if dados is not None
            else None
        )

    @property
    def numero_registros_arquivo_forward(self) -> Optional[int]:
        """
        O número de registros do arquivo forward.

        :return: O número de registros
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return (
            dados.numero_registros_arquivo_forward
            if dados is not None
            else None
        )

    @property
    def numero_registros_necessarios_estagio(self) -> Optional[int]:
        """
        O número de registros necessários por estágio, no arquivo forward.

        :return: O número de registros por estágio
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return (
            dados.numero_registros_necessarios_estagio
            if dados is not None
            else None
        )

    @property
    def ano_inicio_estudo(self) -> Optional[int]:
        """
        O ano de início do estudo, configurado no arquivo `dger.dat`.

        :return: O ano de início do estudo
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.ano_inicio_estudo if dados is not None else None

    @property
    def ano_inicio_historico_vazoes(self) -> Optional[int]:
        """
        O ano de início do histórico de vazões, configurado no arquivo `dger.dat`.

        :return: O ano de início do histórico
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.ano_inicio_historico_vazoes if dados is not None else None

    @property
    def numero_anos_descontar_historico_vazoes(self) -> Optional[int]:
        """
        O número de anos a descontar do histórico.

        :return: O número de anos
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return (
            dados.numero_anos_descontar_historico_vazoes
            if dados is not None
            else None
        )

    @property
    def numero_estagios_ano(self) -> Optional[int]:
        """
        O número de estágios por ano, configurado através da duração do período
        no `dger.dat`.

        :return: O número de estágios por ano
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.numero_estagios_ano if dados is not None else None

    @property
    def mes_inicio_estudo(self) -> Optional[int]:
        """
        O mês de início do estudo, configurado no `dger.dat`.

        :return: O mês de início do período de estudo
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.mes_inicio_estudo if dados is not None else None

    @property
    def mes_inicio_pre_estudo(self) -> Optional[int]:
        """
        O mês de início do pré-estudo, configurado no `dger.dat`.

        :return: O mês de início do período pré-estudo
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.mes_inicio_pre_estudo if dados is not None else None

    @property
    def numero_estagios_pre_estudo(self) -> Optional[int]:
        """
        O número de estágios pré-estudo.

        :return: O número de estágios pré-estudo
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.numero_estagios_pre_estudo if dados is not None else None

    @property
    def numero_patamares_carga(self) -> Optional[int]:
        """
        O número de patamares de carga, configurado no `patamar.dat`

        :return: O número de patamares
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.numero_patamares_carga if dados is not None else None

    @property
    def ordem_maxima_parp(self) -> Optional[int]:
        """
        A ordem máxima dos modelos PAR(p), configurada no `dger.dat`

        :return: A ordem máxima dos modelos
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.ordem_maxima_parp if dados is not None else None

    @property
    def ano_inicio_series_historicas_simuladas(self) -> Optional[List[int]]:
        """
        O ano de início das séries históricas simuladas, caso seja feita
        simulação histórica.

        :return: A ordem máxima dos modelos
        :rtype: list[int] | None
        """
        dados = self.__bloco_dados()
        return (
            dados.ano_inicio_series_historicas_simuladas
            if dados is not None
            else None
        )

    @property
    def numero_anos_historico_vazoes(self) -> Optional[int]:
        """
        O número de anos existentes no histórico de vazões.

        :return: O número de anos
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return (
            dados.numero_anos_historico_vazoes if dados is not None else None
        )

    @property
    def numero_total_submercados(self) -> Optional[int]:
        """
        O número total de submercados, considerando fictícios, cadastrados
        no `sistema.dat`.

        :return: O número total de submercados
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return dados.numero_total_submercados if dados is not None else None

    @property
    def simulacao_final_individualizada(self) -> Optional[int]:
        """
        O uso de simulação final individualizada, configurado no `dger.dat`.

        :return: O valor do flag de simulação final individualizada
        :rtype: int | None
        """
        dados = self.__bloco_dados()
        return (
            dados.simulacao_final_individualizada
            if dados is not None
            else None
        )



from inewave.nwlistop.modelos.blocos.restricao import Restricao
from inewave.nwlistop.modelos.arquivos.arquivorestricao import (
    ArquivoRestricao,
)
from inewave.nwlistop.modelos.c_v_rhv import CVRHVAnos


class CVRHV(ArquivoRestricao):
    """
    Armazena os dados das saídas referentes ao custo de violação
    de RHV por restrição.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `c_v_rhvXXX.out`.

    """

    BLOCKS = [
        Restricao,
        CVRHVAnos,
    ]



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousina import (
    ArquivoUsina,
)
from inewave.nwlistop.modelos.viol_neg_vretiruh import ViolNegVretiruhAnos


class ViolNegVretiruh(ArquivoUsina):
    """
    Armazena os dados das saídas referentes ao desvios negativos da
    retirada de água da usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `viol_neg_vretiruhXXX.out`, onde x varia conforme
    a usina em questão.

    """

    BLOCKS = [
        Usina,
        ViolNegVretiruhAnos,
    ]



from inewave.newave.modelos.caso import NomeCaso, CaminhoGerenciadorProcessos

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional


class Caso(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao caso de estudo.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que podem ser modificadas através do arquivo `caso.dat`.

    """

    T = TypeVar("T")

    SECTIONS = [NomeCaso, CaminhoGerenciadorProcessos]

    @property
    def arquivos(self) -> Optional[str]:
        """
        Caminho para o arquivo `arquivos.dat` de entrada do NEWAVE.

        :return: O caminho para o arquivo
        :rtype: str | None
        """
        b = self.data.get_sections_of_type(NomeCaso)
        if isinstance(b, NomeCaso):
            return b.data
        return None

    @arquivos.setter
    def arquivos(self, a: str):
        b = self.data.get_sections_of_type(NomeCaso)
        if isinstance(b, NomeCaso):
            b.data = a

    @property
    def gerenciador_processos(self) -> Optional[str]:
        """
        Caminho para o gerenciador de processos do NEWAVE.

        :return: O caminho para o arquivo
        :rtype: str | None
        """
        b = self.data.get_sections_of_type(CaminhoGerenciadorProcessos)
        if isinstance(b, CaminhoGerenciadorProcessos):
            return b.data
        return None

    @gerenciador_processos.setter
    def gerenciador_processos(self, a: str):
        b = self.data.get_sections_of_type(CaminhoGerenciadorProcessos)
        if isinstance(b, CaminhoGerenciadorProcessos):
            b.data = a



MockSelcor = [
    "PARAMETROS UTILIZADO NA SELECAO DE CORTES DE BENDERS\n",
    "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX   XXXX  XXXX \n",
    "ITERACAO INICIAL APLICACAO DA SELECAO DE CORTES                0002\n",
    "TAMANHO DA JANELA DE CORTES ATIVOS (k2)                        0000\n",
    "QDE DE CORTES A SEREM ADICIONADOS POR ITERACAO (NADIC)         0012\n",
    "CONSIDERA CORTES DA PROPRIA ITERACAO                           0001\n",
    "IMPRIME RELATORIO DE SAIDA (selcor-xxxx.log)                   0000\n",
    "PERIODO INICIAL E FINAL PARA IMPRESSAO                         0007  0015    (campo 1 vazio ou zero: todo o horizonte)\n",
    "SERITE INICIAL E FINAL PARA IMPRESSAO                          0002  0009    (campo 1 vazio ou zero: todas as series da politica)\n",
]



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)
from inewave.nwlistop.modelos.cviol_rhv_sin import CviolRhvsinAnos


class CviolRhvsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes ao custo de violação
    de RHV para o SIN

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `cviol_rhv_sin.out`.

    """

    BLOCKS = [
        CviolRhvsinAnos,
    ]



from inewave.newave.avl_cortesfpha_nwv import AvlCortesFpha

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.avl_cortesfpha_nwv import (
    MockAvlCortesFphaNwv28,
    MockAvlCortesFphaNwv,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_atributos_encontrados_avl_nwv_cortesfpha_nwv():
    m: MagicMock = mock_open(read_data="".join(MockAvlCortesFphaNwv))
    with patch("builtins.open", m):
        AvlCortesFpha.set_version("28.16")
        rel = AvlCortesFpha.read(ARQ_TESTE)
        assert rel.versao == "28.16_CPAMP"
        assert rel.tabela.at[0, "codigo_usina"] == 4
        assert rel.tabela.at[0, "periodo"] == 5
        assert rel.tabela.at[0, "nome_usina"] == "FUNIL-GRANDE"
        assert rel.tabela.at[0, "indice_corte"] == 1
        assert rel.tabela.at[0, "fator_correcao"] == 1.0
        assert rel.tabela.at[0, "rhs_energia"] == 0.0
        assert rel.tabela.at[0, "coeficiente_volume_util_MW_hm3"] == 0.0
        assert (
            rel.tabela.at[0, "coeficiente_vazao_turbinada_MW_m3s"]
            == 0.35607775
        )
        assert rel.tabela.at[0, "coeficiente_vazao_vertida_MW_m3s"] == 0.0
        assert rel.tabela.at[0, "coeficiente_vazao_lateral_MW_m3s"] == 0.0


def test_atributos_encontrados_avl_nwv_cortesfpha_nwv28():
    m: MagicMock = mock_open(read_data="".join(MockAvlCortesFphaNwv28))
    with patch("builtins.open", m):
        AvlCortesFpha.set_version("28")
        rel = AvlCortesFpha.read(ARQ_TESTE)
        assert rel.versao == "FPHA_NEWAVE"
        assert rel.tabela.at[0, "codigo_usina"] == 4
        assert rel.tabela.at[0, "periodo"] == 1
        assert rel.tabela.at[0, "nome_usina"] == "FUNIL-GRANDE"
        assert rel.tabela.at[0, "indice_corte"] == 1
        assert rel.tabela.at[0, "fator_correcao"] == 1.0
        assert rel.tabela.at[0, "rhs_energia"] == 0.0
        assert rel.tabela.at[0, "coeficiente_volume_util_MW_hm3"] == 0.0
        assert (
            rel.tabela.at[0, "coeficiente_vazao_turbinada_MW_m3s"]
            == 0.35607775
        )
        assert rel.tabela.at[0, "coeficiente_vazao_vertida_MW_m3s"] == 0.0
        assert rel.tabela.at[0, "coeficiente_vazao_lateral_MW_m3s"] == 0.0


def test_eq_avl_nwv_cortesfpha_nwv():
    m: MagicMock = mock_open(read_data="".join(MockAvlCortesFphaNwv))
    with patch("builtins.open", m):
        AvlCortesFpha.set_version("28.16")
        rel1 = AvlCortesFpha.read(ARQ_TESTE)
        rel2 = AvlCortesFpha.read(ARQ_TESTE)
        assert rel1 == rel2


def test_neq_avl_nwv_cortesfpha_nwv():
    m: MagicMock = mock_open(read_data="".join(MockAvlCortesFphaNwv))
    with patch("builtins.open", m):
        AvlCortesFpha.set_version("28.16")
        rel1 = AvlCortesFpha.read(ARQ_TESTE)
        rel2 = AvlCortesFpha.read(ARQ_TESTE)
        rel1.tabela.iloc[0, 0] = -1
        assert rel1 != rel2



# Imports de módulos externos
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField


from inewave.newave.modelos.blocos.tabelacsv import TabelaCSV


class TabelaCortesEvap(TabelaCSV):
    """
    Bloco com as informações dos cortes da evaporação linear.
    """

    __slots__ = []

    BEGIN_PATTERN = "-----;-----;--------------;"
    LINE_MODEL = Line(
        [
            IntegerField(size=5),
            IntegerField(size=5),
            LiteralField(size=14),
            FloatField(size=22, decimal_digits=10),
            FloatField(size=22, decimal_digits=10),
            FloatField(size=11, decimal_digits=2),
            FloatField(size=11, decimal_digits=2),
            FloatField(size=17, decimal_digits=10),
            FloatField(size=12, decimal_digits=14),
        ],
        delimiter=";",
    )
    COLUMN_NAMES = [
        "periodo",
        "codigo_usina",
        "nome_usina",
        "derivada_cota_area",
        "derivada_volume_cota",
        "volume_referencia_hm3",
        "evaporacao_referencia_hm3",
        "coeficiente_volume",
        "rhs_volume",
    ]
    END_PATTERN = ""



from typing import Type, TypeVar, Optional, List, Union
from cfinterface.components.register import Register
from cfinterface.files.registerfile import RegisterFile
from datetime import datetime
import pandas as pd  # type: ignore
from inewave.libs.modelos.usinas_hidreletricas import (
    HidreletricaCurvaJusante,
    HidreletricaCurvaJusantePolinomioPorPartes,
    HidreletricaCurvaJusantePolinomioPorPartesSegmento,
    HidreletricaCurvaJusanteAfogamentoExplicitoUsina,
    HidreletricaCurvaJusanteAfogamentoExplicitoPadrao,
    EstacaoBombeamentoLimitesPeriodoPatamar,
    EstacaoBombeamentoSubmercado,
    EstacaoBombeamento,
    HidreletricaProdutibilidadeEspecificaGrade,
    HidreletricaPerdaHidraulicaGrade,
    VolumeReferencialTipoPadrao,
    VolumeReferencialPeriodo,
)


class UsinasHidreletricas(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes aos dados
    das usinas hidrelétricas do problema.
    """

    T = TypeVar("T", bound=Register)

    REGISTERS = [
        HidreletricaCurvaJusanteAfogamentoExplicitoUsina,
        HidreletricaCurvaJusanteAfogamentoExplicitoPadrao,
        HidreletricaCurvaJusantePolinomioPorPartesSegmento,
        HidreletricaCurvaJusantePolinomioPorPartes,
        HidreletricaCurvaJusante,
        HidreletricaProdutibilidadeEspecificaGrade,
        HidreletricaPerdaHidraulicaGrade,
        EstacaoBombeamentoLimitesPeriodoPatamar,
        EstacaoBombeamentoSubmercado,
        EstacaoBombeamento,
        VolumeReferencialTipoPadrao,
        VolumeReferencialPeriodo,
    ]

    def __registros_ou_df(
        self, t: Type[T], **kwargs
    ) -> Optional[Union[T, List[T], pd.DataFrame]]:
        if kwargs.get("df"):
            return self._as_df(t)
        else:
            kwargs_sem_df = {k: v for k, v in kwargs.items() if k != "df"}
            return self.data.get_registers_of_type(t, **kwargs_sem_df)

    def hidreletrica_curvajusante(
        self,
        codigo_usina: Optional[int] = None,
        indice_familia: Optional[int] = None,
        nivel_montante_referencia: Optional[float] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            HidreletricaCurvaJusante,
            List[HidreletricaCurvaJusante],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que cadastram uma família de curvas
        de jusante para uma usina hidrelétrica. Opcionalmente,
        o retorno pode ser transformado em um `DataFrame`, apenas
        para leitura das informações.

        :param codigo_usina: código que especifica a usina
        :type codigo_usina: int | None
        :param indice_familia: índice da família de polinômios
        :type indice_familia: int | None
        :param nivel_montante_referencia: nível de montante de usina de
            jusante para cálculo da queda
        :type nivel_montante_referencia: float | None
        :param df: ignorar os filtros e retornar
            todos os dados de registros como um DataFrame
        :type df: bool
        :return: Um ou mais registros, se existirem.
        :rtype: `HidreletricaCurvaJusante` |
            List[`HidreletricaCurvaJusante`] | `None` | `DataFrame`
        """
        return self.__registros_ou_df(
            HidreletricaCurvaJusante,
            codigo_usina=codigo_usina,
            indice_familia=indice_familia,
            nivel_montante_referencia=nivel_montante_referencia,
            df=df,
        )

    def hidreletrica_curvajusante_polinomio(
        self,
        codigo_usina: Optional[int] = None,
        indice_familia: Optional[int] = None,
        numero_polinomios: Optional[int] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            HidreletricaCurvaJusantePolinomioPorPartes,
            List[HidreletricaCurvaJusantePolinomioPorPartes],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que cadastram uma família de curvas
        de jusante para uma usina hidrelétrica. Opcionalmente,
        o retorno pode ser transformado em um `DataFrame`, apenas
        para leitura das informações.

        :param codigo_usina: código que especifica a usina
        :type codigo_usina: int | None
        :param indice_familia: índice da família de polinômios
        :type indice_familia: int | None
        :param numero_polinomios: número de polinômios da família
        :type numero_polinomios: int | None
        :param df: ignorar os filtros e retornar
            todos os dados de registros como um DataFrame
        :type df: bool
        :return: Um ou mais registros, se existirem.
        :rtype: `HidreletricaCurvaJusantePolinomioPorPartes` |
            List[`HidreletricaCurvaJusantePolinomioPorPartes`] | `None` | `DataFrame`
        """
        return self.__registros_ou_df(
            HidreletricaCurvaJusantePolinomioPorPartes,
            codigo_usina=codigo_usina,
            indice_familia=indice_familia,
            numero_polinomios=numero_polinomios,
            df=df,
        )

    def hidreletrica_curvajusante_polinomio_segmento(
        self,
        codigo_usina: Optional[int] = None,
        indice_familia: Optional[int] = None,
        indice_polinomio: Optional[int] = None,
        limite_inferior_vazao_jusante: Optional[float] = None,
        limite_superior_vazao_jusante: Optional[float] = None,
        coeficiente_a0: Optional[float] = None,
        coeficiente_a1: Optional[float] = None,
        coeficiente_a2: Optional[float] = None,
        coeficiente_a3: Optional[float] = None,
        coeficiente_a4: Optional[float] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            HidreletricaCurvaJusantePolinomioPorPartesSegmento,
            List[HidreletricaCurvaJusantePolinomioPorPartesSegmento],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que cadastram os polinômios para cada família de curvas
        de jusante para uma usina hidrelétrica. Opcionalmente,
        o retorno pode ser transformado em um `DataFrame`, apenas
        para leitura das informações.

        :param codigo_usina: código que especifica a usina
        :type codigo_usina: int | None
        :param indice_familia: índice da família de polinômios
        :type indice_familia: int | None
        :param indice_polinomio: índice do polinômio da família
        :type indice_polinomio: int | None
        :param limite_inferior_vazao_jusante: limite inferior de vazão de
            jusante para janela de validade do polinômio
        :type limite_inferior_vazao_jusante: float | None
        :param limite_superior_vazao_jusante: limite superior de vazão de
            jusante para janela de validade do polinômio
        :type limite_superior_vazao_jusante: float | None
        :param coeficiente_a0: coeficiente de grau 0 do polinômio
        :type coeficiente_a0: float | None
        :param coeficiente_a1: coeficiente de grau 1 do polinômio
        :type coeficiente_a1: float | None
        :param coeficiente_a2: coeficiente de grau 2 do polinômio
        :type coeficiente_a2: float | None
        :param coeficiente_a3: coeficiente de grau 3 do polinômio
        :type coeficiente_a3: float | None
        :param coeficiente_a4: coeficiente de grau 4 do polinômio
        :type coeficiente_a4: float | None
        :param df: ignorar os filtros e retornar
            todos os dados de registros como um DataFrame
        :type df: bool
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`HidreletricaCurvaJusantePolinomioPorPartesSegmento` |
            list[:class:`HidreletricaCurvaJusantePolinomioPorPartesSegmento`] |
            None
        """
        return self.__registros_ou_df(
            HidreletricaCurvaJusantePolinomioPorPartesSegmento,
            codigo_usina=codigo_usina,
            indice_familia=indice_familia,
            indice_polinomio=indice_polinomio,
            limite_inferior_vazao_jusante=limite_inferior_vazao_jusante,
            limite_superior_vazao_jusante=limite_superior_vazao_jusante,
            coeficiente_a0=coeficiente_a0,
            coeficiente_a1=coeficiente_a1,
            coeficiente_a2=coeficiente_a2,
            coeficiente_a3=coeficiente_a3,
            coeficiente_a4=coeficiente_a4,
            df=df,
        )

    def hidreletrica_curvajusante_afogamentoexplicito_usina(
        self,
        codigo_usina: Optional[int] = None,
        considera_afogamento: Optional[str] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            HidreletricaCurvaJusanteAfogamentoExplicitoUsina,
            List[HidreletricaCurvaJusanteAfogamentoExplicitoUsina],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que habilitam ou desabilitam a consideração
        do tratamento do afogamento explícito por usina. Opcionalmente,
        o retorno pode ser transformado em um `DataFrame`, apenas
        para leitura das informações.

        :param codigo_usina: código que especifica a usina
        :type codigo_usina: int | None
        :param considera_afogamento: habilitação do afogamento
        :type considera_afogamento: str | None
        :type df: bool
        :return: Um ou mais registros, se existirem.
        :rtype: `HidreletricaCurvaJusante` |
            List[`HidreletricaCurvaJusante`] | `None` | `DataFrame`
        """
        return self.__registros_ou_df(
            HidreletricaCurvaJusanteAfogamentoExplicitoUsina,
            codigo_usina=codigo_usina,
            considera_afogamento=considera_afogamento,
            df=df,
        )

    def hidreletrica_curvajusante_afogamentoexplicito_padrao(
        self, considera_afogamento: Optional[str] = None, df: bool = False
    ) -> Optional[
        Union[
            HidreletricaCurvaJusanteAfogamentoExplicitoPadrao,
            List[HidreletricaCurvaJusanteAfogamentoExplicitoPadrao],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que habilitam ou desabilitam a consideração
        do tratamento do afogamento explícito padrão.

        :param considera_afogamento: habilitação do afogamento
        :type considera_afogamento: str | None
        """
        return self.__registros_ou_df(
            HidreletricaCurvaJusanteAfogamentoExplicitoPadrao,
            considera_afogamento=considera_afogamento,
            df=df,
        )

    # -----------

    def hidreletrica_produtibilidade_especifica_grade(
        self, codigo_usina: Optional[int] = None, df: bool = False
    ) -> Optional[
        Union[
            HidreletricaProdutibilidadeEspecificaGrade,
            List[HidreletricaProdutibilidadeEspecificaGrade],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que determinam os valores de produtibilidade
        específica de uma UHE em grade.

        :param codigo_usina: código da usina associada
        :type codigo_usina: int | None
        """
        return self.__registros_ou_df(
            HidreletricaProdutibilidadeEspecificaGrade,
            codigo_usina=codigo_usina,
            df=df,
        )

    def hidreletrica_perda_hidraulica_grade(
        self, codigo_usina: Optional[int] = None, df: bool = False
    ) -> Optional[
        Union[
            HidreletricaPerdaHidraulicaGrade,
            List[HidreletricaPerdaHidraulicaGrade],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que determinam os valores de perda
        hidráulica de uma UHE em grade.

        :param codigo_usina: código da usina associada
        :type codigo_usina: int | None
        """
        return self.__registros_ou_df(
            HidreletricaPerdaHidraulicaGrade,
            codigo_usina=codigo_usina,
            df=df,
        )

    def estacao_bombeamento_limites_periodo_patamar(
        self,
        codigo_estacao: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        patamar: Optional[int] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            EstacaoBombeamentoLimitesPeriodoPatamar,
            List[EstacaoBombeamentoLimitesPeriodoPatamar],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que determinam os limites de bombeamento
        de uma estação por período e patamar.

        :param codigo_estacao: código da estação associada
        :type codigo_estacao: int | None
        :param data_inicio: data de início da validade dos limites
        :type data_inicio: datetime | None
        :param data_fim: data de fim da validade dos limites
        :type data_fim: datetime | None
        :param patamar: patamar de validade dos limites
        :type patamar: int | None
        """
        return self.__registros_ou_df(
            EstacaoBombeamentoLimitesPeriodoPatamar,
            codigo_estacao=codigo_estacao,
            data_inicio=data_inicio,
            data_fim=data_fim,
            patamar=patamar,
            df=df,
        )

    def estacao_bombeamento_submercado(
        self,
        codigo_estacao: Optional[int] = None,
        codigo_submercado: Optional[int] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            EstacaoBombeamentoSubmercado,
            List[EstacaoBombeamentoSubmercado],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que relacionam estações de bombeamento
        e submercados.

        :param codigo_estacao: código da estação associada
        :type codigo_estacao: int | None
        :param codigo_submercado: código do submercado associado
        :type codigo_submercado: int | None
        """
        return self.__registros_ou_df(
            EstacaoBombeamentoSubmercado,
            codigo_estacao=codigo_estacao,
            codigo_submercado=codigo_submercado,
            df=df,
        )

    def estacao_bombeamento(
        self,
        codigo_estacao: Optional[int] = None,
        nome_estacao: Optional[str] = None,
        codigo_usina_origem: Optional[int] = None,
        codigo_usina_destino: Optional[int] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            EstacaoBombeamento,
            List[EstacaoBombeamento],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que definem uma estação de bombeamento.

        :param codigo_estacao: código da estação associada
        :type codigo_estacao: int | None
        """
        return self.__registros_ou_df(
            EstacaoBombeamento,
            codigo_estacao=codigo_estacao,
            nome_estacao=nome_estacao,
            codigo_usina_origem=codigo_usina_origem,
            codigo_usina_destino=codigo_usina_destino,
            df=df,
        )

    def volume_referencial_tipo_padrao(
        self,
        df: bool = False,
    ) -> Optional[
        Union[
            VolumeReferencialTipoPadrao,
            List[VolumeReferencialTipoPadrao],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que definem o tipo de volume de referência.
        """
        return self.__registros_ou_df(
            VolumeReferencialTipoPadrao,
            df=df,
        )

    def volume_referencial_periodo(
        self,
        codigo_usina: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            VolumeReferencialPeriodo,
            List[VolumeReferencialPeriodo],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém registros que definem o volume de referência por perídoo
        para cada usina hidrelétrica.

        :param codigo_usina: código da usina associada
        :type codigo_usina: int | None
        :param data_inicio: data de início da validade da referência
        :type data_inicio: datetime | None
        :param data_fim: data de fim da validade da referência
        :type data_fim: datetime | None
        """
        return self.__registros_ou_df(
            VolumeReferencialPeriodo,
            codigo_usina=codigo_usina,
            data_inicio=data_inicio,
            data_fim=data_fim,
            df=df,
        )



from typing import Optional
from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.datetimefield import DatetimeField
from datetime import datetime


class RegistroPEECadastro(Register):
    """
    Registro que contém um cadastro de um parque eólico equivalente.
    """

    IDENTIFIER = "PEE-CAD"
    IDENTIFIER_DIGITS = 7
    LINE = Line(
        [
            IntegerField(),
            LiteralField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_pee(self) -> Optional[int]:
        return self.data[0]

    @codigo_pee.setter
    def codigo_pee(self, c: int):
        self.data[0] = c

    @property
    def nome_pee(self) -> Optional[str]:
        return self.data[1]

    @nome_pee.setter
    def nome_pee(self, n: str):
        self.data[1] = n


class RegistroPEEPotenciaInstaladaPeriodo(Register):
    """
    Registro que contém a informação de início e fim de
    operação para uma certa potência instalada de um parque
    eólico equivalente.
    """

    IDENTIFIER = "PEE-POT-INST-PER"
    IDENTIFIER_DIGITS = 16
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            FloatField(decimal_digits=3),
        ],
        delimiter=";",
    )

    @property
    def codigo_pee(self) -> Optional[int]:
        """
        O código do PEE.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_pee.setter
    def codigo_pee(self, c: int):
        self.data[0] = c

    @property
    def periodo_inicial(self) -> Optional[datetime]:
        """
        A data de início para a validade da potência instalada.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @periodo_inicial.setter
    def periodo_inicial(self, p: datetime):
        self.data[1] = p

    @property
    def periodo_final(self) -> Optional[datetime]:
        """
        A data de fim para a validade da potência instalada.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @periodo_final.setter
    def periodo_final(self, p: datetime):
        self.data[2] = p

    @property
    def potencia_instalada(self) -> Optional[float]:
        return self.data[3]

    @potencia_instalada.setter
    def potencia_instalada(self, n: float):
        self.data[3] = n


class RegistroPEEConfiguracaoPeriodo(Register):
    """
    Registro que contém um estado de operação de um PEE por
    período.
    """

    IDENTIFIER = "PEE-CONFIG-PER"
    IDENTIFIER_DIGITS = 14
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            LiteralField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_pee(self) -> Optional[int]:
        """
        O código do PEE.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_pee.setter
    def codigo_pee(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início para a validade do estado de operação.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, n: datetime):
        self.data[1] = n

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim para a validade do estado de operação.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, n: datetime):
        self.data[2] = n

    @property
    def estado_operacao(self) -> Optional[str]:
        """
        O estado de operação.

        :return: O estado
        :rtype: Optional[str]
        """
        return self.data[3]

    @estado_operacao.setter
    def estado_operacao(self, n: str):
        self.data[3] = n


class RegistroPEEFTE(Register):
    """
    Registro que contém as informações da função de produção
    linear vento-potência para um PEE.
    """

    IDENTIFIER = "PEE-FPVP-LIN-PU-PER"
    IDENTIFIER_DIGITS = 19
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            FloatField(size=20, decimal_digits=17),
            FloatField(size=20, decimal_digits=17),
        ],
        delimiter=";",
    )

    @property
    def codigo_pee(self) -> Optional[int]:
        """
        O código do PEE.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_pee.setter
    def codigo_pee(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início para a validade da função
        de transformação eólica.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim para a validade da função
        de transformação eólica.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, c: datetime):
        self.data[2] = c

    @property
    def coeficiente_linear(self) -> Optional[float]:
        """
        O coeficiente linear para a FTE

        :return: O coeficiente
        :rtype: Optional[float]
        """
        return self.data[3]

    @coeficiente_linear.setter
    def coeficiente_linear(self, c: float):
        self.data[3] = c

    @property
    def coeficiente_angular(self) -> Optional[float]:
        """
        O coeficiente angular para a FTE

        :return: O coeficiente
        :rtype: Optional[float]
        """
        return self.data[4]

    @coeficiente_angular.setter
    def coeficiente_angular(self, c: float):
        self.data[4] = c


class RegistroPEEGeracaoPatamar(Register):
    """
    Registro que contém as profundidades dos patamares de cada
    período para o histórico de geração de um PEE.
    """

    IDENTIFIER = "PEE-GER-PROF-PER-PAT"
    IDENTIFIER_DIGITS = 20
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            IntegerField(),
            FloatField(decimal_digits=4),
        ],
        delimiter=";",
    )

    @property
    def codigo_pee(self) -> Optional[int]:
        """
        O código do PEE.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_pee.setter
    def codigo_pee(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início para a geração.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim para a geração.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, c: datetime):
        self.data[2] = c

    @property
    def indice_patamar(self) -> Optional[int]:
        """
        O índice do patamar.

        :return: O patamar
        :rtype: Optional[int]
        """
        return self.data[3]

    @indice_patamar.setter
    def indice_patamar(self, c: int):
        self.data[3] = c

    @property
    def profundidade(self) -> Optional[float]:
        """
        O valor de profundidade da geração no patamar.

        :return: A profundidade do patamar
        :rtype: Optional[float]
        """
        return self.data[4]

    @profundidade.setter
    def profundidade(self, c: float):
        self.data[4] = c


class RegistroHistoricoVentoHorizonte(Register):
    """
    Registro que contém o horizonte de históricos de vento.
    """

    IDENTIFIER = "VENTO-HIST-HORIZ"
    IDENTIFIER_DIGITS = 16
    LINE = Line(
        [
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
        ],
        delimiter=";",
    )

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início do horizonte a ser considerado.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[0]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[0] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim do horizonte a ser considerado.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_fim.setter
    def data_fim(self, c: datetime):
        self.data[1] = c


class RegistroHistoricoVento(Register):
    """
    Registro que contém os históricos de vento para o horizonte
    considerado em um PEE.
    """

    IDENTIFIER = "VENTO-HIST"
    IDENTIFIER_DIGITS = 10
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            FloatField(decimal_digits=2),
            FloatField(decimal_digits=2),
        ],
        delimiter=";",
    )

    @property
    def codigo_posto(self) -> Optional[int]:
        """
        O código do posto.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_posto.setter
    def codigo_posto(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início do valor histórico.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim do valor histórico.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, c: datetime):
        self.data[2] = c

    @property
    def velocidade(self) -> Optional[float]:
        """
        A velocidade média do vento no período.

        :return: A velocidade
        :rtype: Optional[float]
        """
        return self.data[3]

    @velocidade.setter
    def velocidade(self, c: float):
        self.data[3] = c

    @property
    def direcao(self) -> Optional[float]:
        """
        A direção média do vento no período.

        :return: A direção
        :rtype: Optional[float]
        """
        return self.data[4]

    @direcao.setter
    def direcao(self, c: float):
        self.data[4] = c


class RegistroPostoVentoCadastro(Register):
    """
    Registro que contém um cadastro de um posto de vento.
    """

    IDENTIFIER = "POSTO-VENTO-CAD"
    IDENTIFIER_DIGITS = 15
    LINE = Line(
        [
            IntegerField(),
            LiteralField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_posto(self) -> Optional[int]:
        """
        O código do posto.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_posto.setter
    def codigo_posto(self, c: int):
        self.data[0] = c

    @property
    def nome_posto(self) -> Optional[str]:
        """
        O nome do posto.

        :return: O nome
        :rtype: Optional[str]
        """
        return self.data[1]

    @nome_posto.setter
    def nome_posto(self, n: str):
        self.data[1] = n


class RegistroPEEPostoVento(Register):
    """
    Registro que contém um mapeamento de PEE para posto de vento
    """

    IDENTIFIER = "PEE-POSTO"
    IDENTIFIER_DIGITS = 9
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_pee(self) -> Optional[int]:
        """
        O código do PEE.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_pee.setter
    def codigo_pee(self, c: int):
        self.data[0] = c

    @property
    def codigo_posto(self) -> Optional[int]:
        """
        O código do posto.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[1]

    @codigo_posto.setter
    def codigo_posto(self, c: int):
        self.data[1] = c


class RegistroPEESubmercado(Register):
    """
    Registro que contém uma relação entre PEE e submercado.
    """

    IDENTIFIER = "PEE-SUBM"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_pee(self) -> Optional[int]:
        """
        O código do PEE.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_pee.setter
    def codigo_pee(self, c: int):
        self.data[0] = c

    @property
    def codigo_submercado(self) -> Optional[int]:
        """
        O código do submercado.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[1]

    @codigo_submercado.setter
    def codigo_submercado(self, c: int):
        self.data[1] = c



from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type, Optional
import pandas as pd  # type: ignore

from inewave.newave.modelos.manutt import BlocoManutencaoUTE


class Manutt(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à programação
    da manutenção das usinas térmicas.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = [BlocoManutencaoUTE]

    @property
    def manutencoes(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as manutenções por usinas.

        - codigo_empresa (`int`)
        - nome_empresa (`str`)
        - codigo_usina (`int`)
        - nome_usina (`str`)
        - codigo_unidade (`int`)
        - data_inicio (`datetime`)
        - duracao (`int`)
        - potencia (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoManutencaoUTE)
        if isinstance(b, BlocoManutencaoUTE):
            return b.data
        return None

    @manutencoes.setter
    def manutencoes(self, valor: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoManutencaoUTE)
        if isinstance(b, BlocoManutencaoUTE):
            b.data = valor
        else:
            raise ValueError("Campo não lido")



from typing import Optional
from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.datetimefield import DatetimeField
from datetime import datetime


class RegistroRE(Register):
    """
    Registro que contém um cadastro de restrição elétrica (RE).
    """

    IDENTIFIER = "RE"
    IDENTIFIER_DIGITS = 2
    LINE = Line(
        [
            IntegerField(),
            LiteralField(size=200),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def formula(self) -> Optional[str]:
        """
        A fórmula da restrição.

        :return: A fórmula
        :rtype: Optional[str]
        """
        return self.data[1]

    @formula.setter
    def formula(self, n: str):
        self.data[1] = n


class RegistroREHorizPer(Register):
    """
    Registro que contém o cadastro do horizonte de validade
    da restrição RE.
    """

    IDENTIFIER = "RE-HORIZ-PER"
    IDENTIFIER_DIGITS = 12
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início da validade da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim da validade da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, n: datetime):
        self.data[2] = n


class RegistroRELimFormPer(Register):
    """
    Registro que contém os limites de cada restrição
    RE por patamar.
    """

    IDENTIFIER = "RE-LIM-FORM-PER"
    IDENTIFIER_DIGITS = 15
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            IntegerField(),
            LiteralField(size=200),
            LiteralField(size=200),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início da validade dos limites da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim da validade dos limites da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, v: datetime):
        self.data[2] = v

    @property
    def patamar(self) -> Optional[int]:
        """
        O índice do patamar para os limites.

        :return: O patamar
        :rtype: Optional[int]
        """
        return self.data[3]

    @patamar.setter
    def patamar(self, v: int):
        self.data[3] = v

    @property
    def limite_inferior(self) -> Optional[str]:
        """
        A equação que da o limite inferior da restrição.

        :return: A equação
        :rtype: Optional[str]
        """
        return self.data[4]

    @limite_inferior.setter
    def limite_inferior(self, v: str):
        self.data[4] = v

    @property
    def limite_superior(self) -> Optional[str]:
        """
        A equação que da o limite superior da restrição.

        :return: A equação
        :rtype: Optional[str]
        """
        return self.data[5]

    @limite_superior.setter
    def limite_superior(self, v: str):
        self.data[5] = v


class RegistroRHE(Register):
    """
    Registro que contém um cadastro de restrição de
    energia (RHE).
    """

    IDENTIFIER = "RHE"
    IDENTIFIER_DIGITS = 3
    LINE = Line(
        [
            IntegerField(),
            LiteralField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def formula(self) -> Optional[str]:
        """
        A fórmula da restrição.

        :return: A fórmula
        :rtype: Optional[str]
        """
        return self.data[1]

    @formula.setter
    def formula(self, n: str):
        self.data[1] = n


class RegistroRHEHorizPer(Register):
    """
    Registro que contém o cadastro do horizonte de validade
    da restrição RHE.
    """

    IDENTIFIER = "RHE-HORIZ-PER"
    IDENTIFIER_DIGITS = 13
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início da validade da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim da validade da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, n: datetime):
        self.data[2] = n


class RegistroRHELsLPPEarmi(Register):
    """
    Registro que contém as retas de cada restrição
    RHE modelada como restrição linear por partes.
    """

    IDENTIFIER = "RHE-LS-LPP-EARMI"
    IDENTIFIER_DIGITS = 16
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
            FloatField(size=16, decimal_digits=4),
            FloatField(size=16, decimal_digits=4),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def indice_reta(self) -> Optional[int]:
        """
        O índice da reta que compõe a restrição.

        :return: O índice da reta
        :rtype: Optional[int]
        """
        return self.data[1]

    @indice_reta.setter
    def indice_reta(self, c: int):
        self.data[1] = c

    @property
    def coeficiente_angular(self) -> Optional[float]:
        """
        O coeficiente angular da reta.

        :return: O coeficiente
        :rtype: Optional[float]
        """
        return self.data[2]

    @coeficiente_angular.setter
    def coeficiente_angular(self, v: float):
        self.data[2] = v

    @property
    def coeficiente_linear(self) -> Optional[float]:
        """
        O coeficiente linear da reta.

        :return: O coeficiente
        :rtype: Optional[float]
        """
        return self.data[3]

    @coeficiente_linear.setter
    def coeficiente_linear(self, v: float):
        self.data[3] = v


class RegistroRHQ(Register):
    """
    Registro que contém um cadastro de restrição de
    vazão (RHQ).
    """

    IDENTIFIER = "RHQ"
    IDENTIFIER_DIGITS = 3
    LINE = Line(
        [
            IntegerField(),
            LiteralField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def formula(self) -> Optional[str]:
        """
        A fórmula da restrição.

        :return: A fórmula
        :rtype: Optional[str]
        """
        return self.data[1]

    @formula.setter
    def formula(self, n: str):
        self.data[1] = n


class RegistroRHQHorizPer(Register):
    """
    Registro que contém o cadastro do horizonte de validade
    da restrição RHQ.
    """

    IDENTIFIER = "RHQ-HORIZ-PER"
    IDENTIFIER_DIGITS = 13
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início da validade da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim da validade da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, n: datetime):
        self.data[2] = n


class RegistroRHQLsLPPVoli(Register):
    """
    Registro que contém as retas de cada restrição
    RHQ modelada como restrição linear por partes.
    """

    IDENTIFIER = "RHQ-LS-LPP-VOLI"
    IDENTIFIER_DIGITS = 15
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
            FloatField(size=16, decimal_digits=4),
            FloatField(size=16, decimal_digits=4),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def indice_reta(self) -> Optional[int]:
        """
        O índice da reta que compõe a restrição.

        :return: O índice da reta
        :rtype: Optional[int]
        """
        return self.data[1]

    @indice_reta.setter
    def indice_reta(self, c: int):
        self.data[1] = c

    @property
    def coeficiente_angular(self) -> Optional[float]:
        """
        O coeficiente angular da reta.

        :return: O coeficiente
        :rtype: Optional[float]
        """
        return self.data[2]

    @coeficiente_angular.setter
    def coeficiente_angular(self, v: float):
        self.data[2] = v

    @property
    def coeficiente_linear(self) -> Optional[float]:
        """
        O coeficiente linear da reta.

        :return: O coeficiente
        :rtype: Optional[float]
        """
        return self.data[3]

    @coeficiente_linear.setter
    def coeficiente_linear(self, v: float):
        self.data[3] = v


class RegistroRHQLimFormPerPat(Register):
    """
    Registro que contém os limites inferiores e superiores
    para restrições RHQ.
    """

    IDENTIFIER = "RHQ-LIM-FORM-PER-PAT"
    IDENTIFIER_DIGITS = 20
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            IntegerField(),
            FloatField(size=16, decimal_digits=2),
            FloatField(size=16, decimal_digits=2),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início da validade dos limites da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim da validade dos limites da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, n: datetime):
        self.data[2] = n

    @property
    def patamar(self) -> Optional[int]:
        """
        O índice do patamar para os limites.

        :return: O patamar.
        :rtype: Optional[int]
        """
        return self.data[3]

    @patamar.setter
    def patamar(self, c: int):
        self.data[3] = c

    @property
    def limite_inferior(self) -> Optional[float]:
        """
        O limite inferior para a restrição.

        :return: O limite inferior
        :rtype: Optional[float]
        """
        return self.data[4]

    @limite_inferior.setter
    def limite_inferior(self, v: float):
        self.data[4] = v

    @property
    def limite_superior(self) -> Optional[float]:
        """
        O limite superior para a restrição.

        :return: O limite superior
        :rtype: Optional[float]
        """
        return self.data[5]

    @limite_superior.setter
    def limite_superior(self, v: float):
        self.data[5] = v


class RegistroRHV(Register):
    """
    Registro que contém um cadastro de restrição de
    volume (RHV).
    """

    IDENTIFIER = "RHV"
    IDENTIFIER_DIGITS = 3
    LINE = Line(
        [
            IntegerField(size=8),
            LiteralField(size=200),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def formula(self) -> Optional[str]:
        """
        A fórmula da restrição.

        :return: A fórmula
        :rtype: Optional[str]
        """
        return self.data[1]

    @formula.setter
    def formula(self, n: str):
        self.data[1] = n


class RegistroRHVHorizPer(Register):
    """
    Registro que contém o cadastro do horizonte de validade
    da restrição RHV.
    """

    IDENTIFIER = "RHV-HORIZ-PER"
    IDENTIFIER_DIGITS = 13
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início da validade da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim da validade da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, n: datetime):
        self.data[2] = n


class RegistroRHVLimFormPer(Register):
    """
    Registro que contém os limites inferiores e superiores
    para restrições RHV.
    """

    IDENTIFIER = "RHV-LIM-FORM-PER"
    IDENTIFIER_DIGITS = 16
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            FloatField(size=16, decimal_digits=2),
            FloatField(size=16, decimal_digits=2),
        ],
        delimiter=";",
    )

    @property
    def codigo_restricao(self) -> Optional[int]:
        """
        O código da restrição.

        :return: O código
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo_restricao.setter
    def codigo_restricao(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início da validade dos limites da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim da validade dos limites da restrição.

        :return: A data
        :rtype: Optional[datetime]
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, n: datetime):
        self.data[2] = n

    @property
    def limite_inferior(self) -> Optional[float]:
        """
        O limite inferior para a restrição.

        :return: O limite inferior
        :rtype: Optional[float]
        """
        return self.data[3]

    @limite_inferior.setter
    def limite_inferior(self, v: float):
        self.data[3] = v

    @property
    def limite_superior(self) -> Optional[float]:
        """
        O limite superior para a restrição.

        :return: O limite superior
        :rtype: Optional[float]
        """
        return self.data[4]

    @limite_superior.setter
    def limite_superior(self, v: float):
        self.data[4] = v



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.cviol_rhq_sin import CviolRhqsinAnos


class CviolRhqsin(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes ao custo de violação
    de RHQ para o SIN

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `cviol_rhq_sin.out`.

    """

    BLOCKS = [
        CviolRhqsinAnos,
    ]



# Rotinas de testes associadas ao arquivo agrint.dat do NEWAVE
from inewave.newave.modelos.agrint import (
    BlocoGruposAgrint,
    BlocoLimitesPorGrupoAgrint,
)

from inewave.newave import Agrint


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.agrint import (
    MockBlocoGruposAgrint,
    MockBlocoLimitesPorGrupoAgrint,
    MockAgrint,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_bloco_grupos_agrint():
    m: MagicMock = mock_open(read_data="".join(MockBlocoGruposAgrint))
    b = BlocoGruposAgrint()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, 0] == 14
    assert b.data.iloc[-1, 2] == 1


def test_bloco_limites_agrint():
    m: MagicMock = mock_open(read_data="".join(MockBlocoLimitesPorGrupoAgrint))
    b = BlocoLimitesPorGrupoAgrint()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, 0] == 14
    assert b.data.iloc[-1, -1] == 17923


def test_atributos_encontrados_agrint():
    m: MagicMock = mock_open(read_data="".join(MockAgrint))
    with patch("builtins.open", m):
        ad = Agrint.read(ARQ_TESTE)
        assert ad.agrupamentos is not None
        assert ad.limites_agrupamentos is not None


def test_atributos_nao_encontrados_agrint():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Agrint.read(ARQ_TESTE)
        assert ad.agrupamentos is None
        assert ad.limites_agrupamentos is None


def test_eq_agrint():
    m: MagicMock = mock_open(read_data="".join(MockAgrint))
    with patch("builtins.open", m):
        cf1 = Agrint.read(ARQ_TESTE)
        cf2 = Agrint.read(ARQ_TESTE)
        assert cf1 == cf2


def test_neq_agrint():
    m: MagicMock = mock_open(read_data="".join(MockAgrint))
    with patch("builtins.open", m):
        cf1 = Agrint.read(ARQ_TESTE)
        cf2 = Agrint.read(ARQ_TESTE)
        cf2.agrupamentos.loc[0, 0] = 0
        assert cf1 != cf2


def test_leitura_escrita_agrint():
    m_leitura: MagicMock = mock_open(read_data="".join(MockAgrint))
    with patch("builtins.open", m_leitura):
        cf1 = Agrint.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Agrint.read(ARQ_TESTE)
        assert cf1 == cf2



from inewave.nwlistop.modelos.blocos.restricao import Restricao
from inewave.nwlistop.modelos.arquivos.arquivorestricaopatamar import (
    ArquivoRestricaoPatamar,
)
from inewave.nwlistop.modelos.viol_rhq import ViolRHQAnos


class ViolRHQ(ArquivoRestricaoPatamar):
    """
    Armazena os dados das saídas referentes aos valores das restrições
    RHQ por restrição.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `viol_rhqXXX.out`.

    """

    BLOCKS = [
        Restricao,
        ViolRHQAnos,
    ]
